let make_arr dim =
  let arr = Array.make dim (Array.make dim ' ') in
  for i = 0 to dim - 1 do arr.(i) <- Array.make dim ' ' done;
  arr

let array2_dim arr =
  (Array.length arr, Array.length arr.(0))

let is_on_grid arr (row, col) =
  let (rows, cols) = array2_dim arr in
  if row < 0 || row >= rows || col < 0 || col >= cols then false
  else true
  
let is_cell arr (row, col) =
  if not @@ is_on_grid arr (row, col) then false
  else arr.(row).(col) != ' '

let set arr (row, col) char =
  if is_on_grid arr (row, col) then
    arr.(row).(col) <- char

let ( ++ ) (a, b) (c, d) = (a + c, b + d)

let empty_neighbors arr (row, col) =
  let (rows, cols) = array2_dim arr in
  let result = ref [] in
  let empty (r, c) = not @@ is_cell arr (r, c) in
  if row > 0 && empty (row - 1, col) then result := (row - 1, col) :: !result;
  if row < rows - 1 && empty (row + 1, col) then result := (row + 1, col) :: !result;
  if col > 0 && empty (row, col - 1) then result := (row, col - 1) :: !result;
  if col < cols - 1 && empty (row, col + 1) then result := (row, col + 1) :: !result;
  !result

let random_neighbor arr (row, col) =
  let ls = [(row + 1, col); (row - 1, col); (row, col + 1); (row, col - 1)] in
  let ls = List.filter (fun x -> is_on_grid arr x) ls in
  let r = Random.int (List.length ls) in
  List.nth ls r
  
let random_empty_neighbor arr (row, col) =
  let candidates = empty_neighbors arr (row, col) in
  let len = List.length candidates in
  if len = 0 then None
  else
    let r = Random.int len in
    Some (List.nth candidates r)
  
let rec draw_path arr (row, col) counter =
  if counter > 0 then
    begin
      let (row, col) =
        match random_empty_neighbor arr (row, col) with
        | None -> random_neighbor arr (row, col)
        | Some (row, col) -> (row, col) in
      set arr (row, col) '+';
      draw_path arr (row, col) (counter - 1)
    end

let place_eggs_boosters arr row_min =
  for row = row_min to Array.length arr - 1 do
    for col = 0 to Array.length arr.(row) - 1 do
      if is_cell arr (row, col) then
        let f delta char =
          if not (is_cell arr ((row, col) ++ delta)) && Random.int 128 < 2 then
            set arr (row, col) char in
        begin
          f (1, 0) 'v';
          f (-1, 0) '^';
          f (0, 1) '>';
          f (0, -1) '<';
          if Random.int 256 = 0 then
            set arr (row, col) 'B';
        end
    done
  done

let main () =
  Random.init 9;
  (* Random.self_init(); *)
  let arr_low = make_arr 100 in
  draw_path arr_low (0, 0) 5000;
  place_eggs_boosters arr_low 0;
  let stream = open_out "voxelworld0.vox" in
  for row = 0 to Array.length arr_low - 1 do
    for col = 0 to Array.length arr_low.(row) - 1 do
      output_char stream arr_low.(row).(col)
    done;
    output_char stream '\n';
  done;
  output_string stream "#ply\n";
  let down = 40 in
  for row = 0 to down do
    for col = 0 to Array.length arr_low.(row) - 1 do
      let c = match arr_low.(row).(col) with ' ' -> ' ' | _ -> '+' in
      output_char stream c
    done;
    output_char stream '\n';
  done;
  for i = 0 to 5 do output_char stream '\n' done;
  let arr = make_arr 100 in
  draw_path arr (0, 0) 5000;
  place_eggs_boosters arr 56;
  for row = 0 to Array.length arr - 1 do
    for col = 0 to Array.length arr.(row) - 1 do
      output_char stream arr.(row).(col)
    done;
    output_char stream '\n';
  done;
  output_char stream '\n';
  close_out stream

let () = main ()
